#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""The Damona configuration"""
import os
import pathlib

import colorlog
from easydev import CustomConfig

import damona.shell

logger = colorlog.getLogger(__name__)


__all__ = ["Config", "get_damona_commands"]


class Config:
    """A place holder to store our configuration file and shell scripts


    This class is called each time damona is started. The config file, if not present
    is created, otherwise nothing happens. Same for the bash and fish shell configuration
    files

    The damona configuration file looks like::

        [general]
        quiet=False

        [zenodo]
        token=APmm6p....
        orcid=0000-0001
        name='Cokelaer, Thomas'
        affiliation='Institut Pasteur'

        [sandbox.zenodo]
        token=FFmbAEhQbb...
        orcid=0000-0001
        name='Cokelaer, Thomas'
        affiliation='Institut Pasteur'

    Where the urls section can be used to store aliases to external registry. When
    installing software using::

        damona install example --url damona

    if the alias damona is in the [urls] section, it is replaced by its real value (https://...)
    the URL must end with the expected registry name **registry.txt**

    The zenodo section is not save by default since it is for developpers only.


    """

    def __init__(self, name="damona", urls={"damona": "https://biomics.pasteur.fr/salsa/damona/registry.txt"}):
        configuration = CustomConfig(f"{name}", verbose=True)

        #  let us add a damona.cfg in it. This will store URLs to look for singularities
        # This is done only once to not overwrite user options
        self.user_config_dir = pathlib.Path(configuration.user_config_dir)

        self.config_file = self.user_config_dir / "damona.cfg"
        if self.config_file.exists() is False:  # pragma: no cover
            with open(self.config_file, "w") as fout:
                fout.write("[general]\n")
                fout.write("verbose=True\n\n")
                fout.write("[urls]\n")
                for k, v in urls.items():
                    fout.write("{}={}".format(k, v))
        else:
            logger.debug("damona.cfg file exists already. Reading it")

        # read the config
        self.read()

        # create the shell script once for all
        created = self.add_bash()
        created = self.add_fish()
        created = self.add_zsh()

        if created:  # pragma: no cover
            logger.critical(
                "Please start a new shell to benefit from " "the configuration file and activate/deactivate command"
            )
            # sys.exit(1)

    def read(self):
        """Reads the config file"""
        from configparser import ConfigParser

        config = ConfigParser()
        config.read_file(open(self.config_file))
        self.config = config

    def add_bash(self):  # pragma: no cover
        if os.path.exists(self.user_config_dir / "damona.sh") is False:
            logger.info("adding a damona.sh in your DAMONA_PATH")
            _damona_config_path = self.user_config_dir
            logger.warning(f"Creating damona.sh file in {_damona_config_path}. ")

            shell_path = pathlib.Path(damona.shell.__path__[0])
            with open(shell_path / "bash" / "damona.sh", "r") as fin:
                with open(_damona_config_path / "damona.sh", "w") as fout:
                    fout.write(fin.read())
            return True
        else:
            return False

    def add_zsh(self):  # pragma: no cover
        if os.path.exists(self.user_config_dir / "damona.zsh") is False:
            logger.info("adding a damona.zsh in your DAMONA_PATH")
            _damona_config_path = self.user_config_dir
            logger.warning(f"Creating damona.zsh file in {_damona_config_path}. ")

            shell_path = pathlib.Path(damona.shell.__path__[0])
            with open(shell_path / "zsh" / "damona.zsh", "r") as fin:
                with open(_damona_config_path / "damona.zsh", "w") as fout:
                    fout.write(fin.read())
            return True
        else:
            return False

    def add_fish(self):
        if os.path.exists(self.user_config_dir / "damona.fish") is False:
            logger.info("adding a damona.fish in your DAMONA_PATH")
            _damona_config_path = self.user_config_dir
            logger.warning(f"Creating damona.fish file in {_damona_config_path}. ")

            shell_path = pathlib.Path(damona.shell.__path__[0])
            with open(shell_path / "fish" / "damona.fish", "r") as fin:
                with open(_damona_config_path / "damona.fish", "w") as fout:
                    fout.write(fin.read())
            return True
        else:
            return False


def get_damona_commands():
    """Print commands available in Damona if not hidden.

    This function is used for the fish completion"""
    from damona import script

    commands = [x for x in dir(script) if "allow_interspersed_args" in dir(getattr(script, x))]
    commands = [x for x in commands if not getattr(script, x).hidden]
    commands = [x for x in commands if x != "main"]
    print("\n".join(commands))



#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
import importlib.metadata as metadata
import os

import colorlog


def get_package_version(package_name):
    try:
        version = metadata.version(package_name)
        return version
    except metadata.PackageNotFoundError:
        return f"{package_name} not found"


version = get_package_version("damona")


# The logger mechanism is here:
handler = colorlog.StreamHandler()
handler.setFormatter(colorlog.ColoredFormatter("%(log_color)s%(levelname)s: [%(name)s,l %(lineno)s]: %(message)s"))
logger = colorlog.getLogger("damona")
logger.addHandler(handler)

# Here we create a persistent config directory in the Home of the user.
# This is a small file.
from damona.config import Config

try:
    Config()
except:  # pragma: no cover
    logger.warning("Could not create a persistent config file in your home. Unexpected error.")


# the following statement checks existence of environemental variables (DAMONA_PATH, DAMONA_EXE, etc)
from damona.common import DamonaInit

DamonaInit()


from damona.common import Damona
from damona.environ import Environ, Environment

# The user/developer API
from damona.registry import Registry



def check_for_updates(package_name, current_version, timeout=2):
    # local import

    import requests
    import json
    from packaging import version

    # Get the current version of the installed package

    # Fetch the latest version from PyPI
    url = f"https://pypi.org/pypi/{package_name}/json"

    try:
        response = requests.get(url, timeout=timeout)
        response.raise_for_status()  # Raise an error for bad status codes (e.g. 404)
        data = response.json()
        latest_version = data["info"]["version"]

        # Compare versions and notify the user
        if version.parse(latest_version) > version.parse(current_version):
            logger.warning(f"\u26a0\ufe0f A new version ({latest_version}) of {package_name} is available! You have {current_version}. Use 'pip install damona --upgrade'")
        else:
            logger.info(f"\u2705 You are using the latest version of {package_name} ({current_version}).")
    except requests.ConnectionError:
        logger.warning("No internet connection. Unable to check for updates.")
    except requests.Timeout:
        logger.warning("The request to check for updates timed out.")
    except requests.RequestException as e:
        logger.error(f"An error occurred: {e} while checking for updates")



check_for_updates("damona", version)






#!/usr/bin/env python

""" MultiQC module to parse output from sequana"""
import os
import re
import json

from collections import OrderedDict

# prevent boring warning (version 1.0)
import logging

logging.captureWarnings(True)
from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table, heatmap, bargraph

logging.captureWarnings(False)

import colorlog

log = colorlog.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """

    Here keys are the sample name

    {
     "25": {
      "Bacteria": 0.4,
      "Metazoa": 2.0,
      "Unclassified": 0.0,
      "Viruses": 97.6
     },
     "26": {
      "Bacteria": 8.800000000000002,
      "Metazoa": 0.4,
      "Unclassified": 18.4,
      "Viruses": 72.39999999999998
     },
    }

    .. warning:: this is not really a multiqc here because we read a file
        that summarizes several samples together.

    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/Kraken",  # name that appears at the top
            anchor="sequana_kraken",  #
            target="sequana",  # Name show that link to the following href
            href="http://github.com/sequana/sequana/",
            info="sequana_kraken pipeline multi summary",
        )

        self.sequana_data = {}

        # In theory only one file
        for myfile in self.find_log_files("sequana_kraken"):
            name = myfile["s_name"]
            d = json.loads(myfile["f"])

            for k, v in d.items():
                S = sum(list(v.values()))
                U = v["Unclassified"]
                v["Classified"] = S - U
                self.sequana_data[k] = v

        if len(self.sequana_data) == 0:
            log.debug("No samples found: sequana_kraken")
            raise UserWarning

        """info = "<ul>"
        for this in sorted(self.sequana_data.keys()):
            info += '<li><a href="{}/summary.html">{}</a></li>'.format(this,this,this)
        info += "</ul>"
        href="http://sequana.readthedocs.io/en/main/"
        target = "Sequana"
        mname = '<a href="{}" target="_blank">{}</a> individual report pages:'.format(href, target)
        self.intro = '<p>{} {}</p>'.format( mname, info)
        """

        log.info("Found {} reports".format(len(self.sequana_data)))

        self.populate_columns()
        self.add_kraken()
        self.add_classified_vs_unclassified()

    def add_classified_vs_unclassified(self):
        data = {}
        for sample_name in self.sequana_data.keys():
            C = self.sequana_data[sample_name]["Classified"]
            U = self.sequana_data[sample_name]["Unclassified"]
            data[sample_name] = {"unclassified": U, "classified": C}

        pconfig = {
            "title": "classification",
            "cpswitch": False,
            "min": 0,
            "max": 100,
            "format": "{0:.2f}",
            "logswitch": False,
        }

        keys = OrderedDict()
        keys["classified"] = {"color": "green", "name": "Classified"}
        keys["unclassified"] = {"color": "red", "name": "Unclassified"}

        self.add_section(
            name="Classification ratio",
            anchor="classification_ratio",
            description="The following barplots shows proportion of classified and unclassified reads in percentage",
            helptext="",
            plot=bargraph.plot(data, keys, pconfig),
        )

    def _set_nan_to_zero(self, x):
        try:
            x + 0
            return x
        except:
            return 0

    def add_kraken(self):
        data = {}

        # First, we figure out all possible names
        kingdoms = set([x for k in self.sequana_data.keys() for x in self.sequana_data[k].keys()])

        colors = ["Archaea", "Bacteria", "Eukaryota", "Viruses", "Metazoa", "Fungi", "Unclassified", "Classified"]

        for sample_name in self.sequana_data.keys():
            for kingdom in sorted(kingdoms):
                if kingdom not in self.sequana_data[sample_name]:
                    self.sequana_data[sample_name][kingdom] = 0

            data[sample_name] = {"others": 0}
            for kingdom in sorted(kingdoms):
                if kingdom not in colors:
                    # here we add together non-superkingdom + other artifical
                    # sequences
                    data[sample_name]["others"] += self._set_nan_to_zero(self.sequana_data[sample_name][kingdom])
                else:
                    data[sample_name][kingdom.lower()] = self._set_nan_to_zero(self.sequana_data[sample_name][kingdom])
            data[sample_name]["unclassified"] = self._set_nan_to_zero(self.sequana_data[sample_name]["Unclassified"])

        pconfig = {
            "title": "Taxonomy by kingdom",
            # "percentages": True,
            "cpswitch": False,
            "min": 0,
            "max": 100,
            "format": "{0:.2f}",
            "logswitch": False,
        }

        keys = OrderedDict()
        # superkingdom:
        keys["archea"] = {"color": "orange", "name": "Archea"}
        keys["bacteria"] = {"color": "#b1084c", "name": "Bacteria"}
        keys["eukaryota"] = {"color": "green", "name": "Eukaryota"}
        keys["viruses"] = {"color": "#437bb1", "name": "Viruses"}
        # kingdom:
        keys["metazoa"] = {"color": "green", "name": "Metazoa"}
        keys["fungi"] = {"color": "purple", "name": "Fungi"}
        # others
        keys["unclassified"] = {"color": "grey", "name": "Unclassified"}
        keys["others"] = {"color": "blue", "name": "Others"}
        # subkingdom
        # keys['viridiplantae'] = {'color': 'yellow', 'name': 'Viridiplantae'}
        # keys['dikarya'] = {'color': 'brown', 'name': 'dikarya'}

        self.add_section(
            name="Taxonomy by kingdom",
            anchor="taxonomy",
            description="The following barplots summarizes the kraken analysis for each sample. ",
            helptext="",
            plot=bargraph.plot(data, keys, pconfig),
        )

    def populate_columns(self):
        headers = {}

        if any(["Classified" in self.sequana_data[s] for s in self.sequana_data]):
            headers["Classified"] = {
                "title": "Classified reads (%)",
                "description": "classified reads (%)",
                "min": 0,
                "max": 100,
                "scale": "RdYlGn",
                "format": "{0:.2f}",
                "shared_key": "count",
            }

        for name in ["Viruses", "Bacteria", "Eukaryota", "Archea", "Fungi"]:
            if any([name in self.sequana_data[s] for s in self.sequana_data]):
                headers[name] = {
                    "title": "Reads classified as {} (%)".format(name.lower()),
                    "description": "Reads classified as {} (%)".format(name.lower()),
                    "min": 0,
                    "max": 100,
                    "scale": "RdYlGn",
                    "format": "{0:.2f}",
                    "shared_key": "count",
                }

        if len(headers.keys()):
            self.general_stats_addcols(self.sequana_data, headers)



#
#  This file is part of Damona software
#
#  Copyright (c) 2020-2021 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Provide some stats for admin"""
import os
import sys

import colorlog

from damona import Registry, version

logger = colorlog.getLogger(__name__)


__all__ = ["stats"]


def stats(biocontainers=False):
    """Prints statistics about Damona

    It includes the number of software and their releases.
    It also includes the nmbr of binaries.

    ::

        from damona.admin import stats
        stats()

    """

    data = {}
    print(f"Damona version : {version}")
    r = Registry(biocontainers=biocontainers)

    Nv = len(r.get_list())
    Nb = len(set([y for x in r.get_binaries().values() for y in x]))
    Ns = len(set([x.split(":")[0] for x in r.get_list()]))
    if biocontainers:
        print("From biocontainers, in theory*, we also have:")
    print(f"- number of containers:  {Ns}")
    print(f"- version: {Nv}")
    print(f"- unique binaries: {Nb}")
    data["version"] = Nv
    data["software"] = Ns
    data["unique_binaries"] = Nb
    if biocontainers:
        print(
            """*: not all software provided in the biocontainers registry are actually on Docker.
There is nothing we can do about that in Damona. Actual number is more around 1000 public software"""
        )

    return data


def get_software_names():
    r = Registry(biocontainers=False)
    return set([x.split(":")[0] for x in r.get_list()])


def build_biocontainers_registry(output="biocontainers.yml", force=False, limit=20000):  # pragma: no cover
    """Create the list of software and their versions available in Biocontainer"""
    logger.info("Retrieve all information from Biocontainers")
    try:
        from bioservices import Biocontainers
    except (ModuleNotFoundError, ImportError) as err:
        logger.error(
            "This function is for admin only. You may use but you mut install bioservices first (pip install bioservices). "
        )
        return

    b = Biocontainers()
    info = b.get_tools(limit=limit)
    if len(info) > limit:
        logger.warning(f"Looks like you reached the limit of {limit} tool. Use limit argument to get more")
    tools = {}
    for name, versions in zip(info["name"], info["versions"]):
        tools[name] = [x["meta_version"] for x in versions]

    # Create the registry
    if os.path.exists(output) and not force:
        logger.error(f"Output file {output} exists already. Please rename or remove the target file")
        sys.exit(1)

    with open(output, "w") as fout:
        for name, versions in tools.items():
            fout.write(f"{name}:\n  releases:\n")
            for version in versions:
                fout.write(f"    {version}:\n")
                fout.write(f"      download: docker://biocontainers/{name}:{version}\n")



#!/usr/bin/env python

""" MultiQC module to parse output from sequana (coverage)"""
import os
import re
from math import log10
from pathlib import Path

# prevent boring warning (version 1.0)
import logging

logging.captureWarnings(True)
from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table, heatmap, bargraph

logging.captureWarnings(False)

import colorlog

log = colorlog.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/coverage",  # name that appears at the top
            anchor="sequana_coverage",
            target="sequana",  # Name show that link to the following href
            href="http://github.com/sequana/sequana/",
            info="sequana_coverage multi Summary",
        )

        self.sequana_data = dict()
        self.sequana_desc = {}
        self.sequana_root = {}

        # Do we have a unique chromosome/contig name ?
        chrom_names = set([])
        for myfile in self.find_log_files("sequana_coverage"):
            name = myfile["s_name"]
            if name.startswith("summary_"):
                name = name.replace("summary_", "")
            data = self.parse_logs(myfile["f"])
            chrom_names.add(data["data"]["chrom_name"])

        self._maps = {}
        if len(chrom_names) == 1:
            self.unique = True
        else:
            self.unique = False

        for myfile in self.find_log_files("sequana_coverage"):
            name = myfile["s_name"]
            if name.startswith("summary_"):
                name = name.replace("summary_", "")
            data = self.parse_logs(myfile["f"])
            if self.unique is True:
                key = data["sample_name"]
            else:
                # convert into string in case
                key = data["sample_name"] + "/" + str(data["data"]["chrom_name"])

            # sometimes keys/name of chromosomes are just integer, so we need to
            # convert to string
            self._maps[key] = data["sample_name"] + "/" + str(data["data"]["chrom_name"])

            self.sequana_data[key] = data["data"]
            self.sequana_desc[key] = data["data_description"]
            self.sequana_root[key] = myfile["root"]
        if len(self.sequana_data) == 0:
            log.debug("No samples found: sequana_coverage")
            raise UserWarning

        info = "<ul>"
        for this in sorted(self.sequana_data.keys()):
            sample_name, chrom_name = self._maps[this].split("/")

            root = self.sequana_root[this]

            # sequana_coverage stores multiqc locally
            # the pipelines stores it in ./multiqc/. We need to handle both cases here

            p1 = Path(config.working_dir).absolute()
            p2 = Path(config.output_dir).absolute()

            if len(str(p1)) > len(str(p2)):
                pp = p1.relative_to(p2)
            else:
                pp = p2.relative_to(p1)

            if len(pp.parts):
                subpath = "/".join([".." for x in pp.parts])

                if root.startswith("./"):
                    root = root[1:]
            else:
                subpath = ""

            info += '<li><a href="{}/{}.cov.html">{}</a></li>'.format(subpath + root, chrom_name, this)
        info += "</ul>"
        href = "http://sequana.readthedocs.io/en/main/"
        target = "Sequana"
        mname = '<a href="{}" target="_blank">{}</a> individual report pages:'.format(href, target)
        self.intro = "<p>{} {}</p>".format(mname, info)

        log.info("Found {} reports".format(len(self.sequana_data)))

        self.populate_columns()
        self.add_hist_coverage()
        self.add_DOC()
        self.add_BOC()
        try:
            self.add_CV()
        except:
            pass
        self.add_length()
        self.add_ROI()
        self.add_C3()

    def parse_logs(self, log_dict):
        import json

        log_dict = json.loads(log_dict)
        return log_dict

    def add_BOC(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"BOC": self.sequana_data[name]["BOC"]}

        pconfig = {
            "title": self.sequana_desc[name]["BOC"],
            "percentages": False,
            "min": 0,
            "max": 0,
            "logswitch": False,
        }
        self.add_section(
            name="Breadth of coverage",
            anchor="boc",
            description="Breadth of coverage: proportion of the genome " + "covered by at least one read",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_DOC(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"DOC": self.sequana_data[name]["DOC"]}

        pconfig = {"title": self.sequana_desc[name]["DOC"], "percentages": False, "min": 100, "logswitch": True}
        self.add_section(
            name="Depth of coverage",
            anchor="doc",
            description="Depth of coverage: average number of reads" + " mapping on each genome position",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_CV(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"CV": self.sequana_data[name]["CV"]}

        pconfig = {"title": self.sequana_desc[name]["CV"], "percentages": False, "min": 100, "logswitch": False}

        self.add_section(
            name="Coefficient of Variation",
            anchor="cv",
            description="The ratio of DOC mean by DOC standard deviation",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_ROI(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"ROI": self.sequana_data[name]["ROI"]}

        pconfig = {"title": self.sequana_desc[name]["ROI"], "percentages": False, "min": 100, "logswitch": False}

        self.add_section(
            name="ROI",
            anchor="ROI",
            description="Number of regions of interest",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_C3(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"C3": self.sequana_data[name]["C3"]}

        pconfig = {"title": self.sequana_desc[name]["C3"], "percentages": False, "min": 0, "max": 0, "logswitch": False}

        self.add_section(
            name="C3",
            anchor="C3",
            description="Centralness (roughly speaking, ratio of " + "outliers versus total genome length).",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_length(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"length": self.sequana_data[name]["length"]}

        pconfig = {"title": self.sequana_desc[name]["length"], "percentages": False, "min": 0, "logswitch": False}
        self.add_section(
            name="Contig length",
            anchor="length",
            description="Length of the contig/chromosome",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_hist_coverage(self):
        data = dict()
        data_norm = dict()
        for s_name in self.sequana_data:
            try:
                X = self.sequana_data[s_name]["hist_coverage"]["X"]
                Y = self.sequana_data[s_name]["hist_coverage"]["Y"]
                Y = [y for y in Y]
                data[s_name] = {x: y if y else 0 for x, y in zip(X, Y)}
            except KeyError:
                pass

        if len(data) == 0:
            log.debug("no data for the coverage plots")
            return None

        pconfig = {
            "id": "sequana_coverage_hist",
            "title": "Depth of Coverage",
            "ylab": "#",
            "xlab": "DOC",
            "ymin": 0,
            #'ymax': 0.2,
            #'xmax': 100,
            "xmin": 0,
            "yDecimals": True,
            "tt_label": "<b>{point.x:.2f} DOC</b>: {point.y:.4f}",
            #'colors': self.get_status_cols('per_sequence_gc_content'),
            "data_labels": [
                # {'name': 'Percentages', 'ylab': 'Percentage'},
                {"name": "Counts", "ylab": "PDF"}
            ],
        }

        self.add_section(
            name="Depth of Coverage Histogram",
            anchor="coverage_hist",
            description=(
                "Histogram (normalised) of the depth of coverage. For"
                " convenience, only  99% the data (centered) to "
                "avoid outliers. For detailled histograms, please see "
                " the links above "
            ),
            # plot = linegraph.plot([data_norm, data], pconfig))
            plot=linegraph.plot(data, pconfig),
        )

    def populate_columns(self):
        headers = {}
        formats = {
            "BOC": "{:,.2f}",
            "CV": "{:,.2f}",
            "DOC": "{:,.2f}",
            "length": "{:,d}",
            "ROI": "{:,d}",
            "C3": "{:,.2f}",
        }

        for field in ["BOC", "DOC", "ROI", "length", "CV", "C3"]:
            # description are supposed to be the same for all samples, let us
            # take the first one
            desc = [self.sequana_desc[s][field] for s in self.sequana_desc][0]

            if any([field in self.sequana_data[s] for s in self.sequana_data]):
                headers[field] = {
                    "title": field,
                    "description": desc,
                    "min": 0,
                    "max": 0,
                    "scale": "RdYlGn",
                    "format": formats[field],
                    "shared_key": "count",
                }

                if field in ["BOC"]:
                    headers[field]["min"] = 0
                    headers[field]["max"] = 100

        if len(headers.keys()):
            self.general_stats_addcols(self.sequana_data, headers)



#!/usr/bin/env python

""" MultiQC module to parse output from sequana"""
import os
import re
import json

from collections import OrderedDict

# prevent boring warning (version 1.0)
import logging

logging.captureWarnings(True)
from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table, heatmap, bargraph

logging.captureWarnings(False)

# Initialise the logger
import colorlog

log = colorlog.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """ """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/laa",  # name that appears at the top
            anchor="sequana_laa",  #
            target="sequana",  # Name show that link to the following href
            href="http://github.com/sequana/sequana/",
            info="amplicon pipeline multi summary",
        )

        self.sequana_data = {}

        # temporary test otherwise multiqc is failing. do not know why for now
        try:
            [x for x in self.find_log_files("sequana_laa")]
        except KeyError:
            return

        # In theory only one file
        for myfile in self.find_log_files("sequana_laa"):
            logging.info("Parsing {}".format(myfile))
            # name = myfile['s_name']
            # name = name.replace("sequana_laa_", "")
            # self.sequana_data[name] = self.parse_logs(myfile["f"])

        if len(self.sequana_data) == 0:
            log.debug("No samples found: sequana_laa")
            raise UserWarning

        log.info("Found {} reports".format(len(self.sequana_data)))

        self.populate_columns()
        self.add_ccs_reads()
        self.add_ccs_reads_hist()

    def parse_logs(self, log_dict):
        import json

        log_dict = json.loads(log_dict)
        return log_dict

    def add_ccs_reads(self):
        data = {}

        for name in self.sequana_data.keys():
            data[name] = {"ccs_reads": self.sequana_data[name]["ccs_reads"]}

        pconfig = {
            "title": "CCS reads",
            "percentages": False,
            "min": 0,
            # "max":100,
            "format": "{0:.2f}",
            "logswitch": False,
        }

        self.add_section(
            name="CCS reads",
            anchor="ccs_reads",
            description="The following barplots summarizes the number of CCS reads.",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_ccs_reads_hist(self):
        data = {}
        # for name in self.sequana_data:
        #    data = [item['ccs_reads'] for item in self.sequana_data.items()]

        data = [self.sequana_data[key]["ccs_reads"] for key in self.sequana_data.keys()]
        from pylab import hist

        Y, X, _ = hist(data, bins=20)

        data = {}
        data["sample"] = {x: y for x, y in zip(X, Y)}

        pconfig = {
            "title": "CCS reads ",
            "percentages": False,
            "min": 0,
            # "max":100,
            "format": "{0:.2f}",
        }

        self.add_section(
            name="CCS reads histogram",
            anchor="ccs_reads_hist",
            description="CCS reads histogram.",
            helptext="",
            plot=linegraph.plot(data, pconfig),
        )

    def populate_columns(self):
        headers = {}

        if any(["ccs_reads" in self.sequana_data[s] for s in self.sequana_data]):
            headers["ccs_reads"] = {
                "title": "CCS reads ",
                "description": "CCS reads",
                "min": 0,
                #'max': 100,
                "scale": "RdYlGn",
                "format": "{0:.2f}",
                "shared_key": "count",
            }

        if len(headers.keys()):
            self.general_stats_addcols(self.sequana_data, headers)



from multiqc.utils import config


def load_config():
    my_search_patterns = {
        "sequana_coverage": {"fn": "sequana_summary_coverage.json"},
        "sequana_pacbio_qc": {"fn": "sequana_summary*.json"},
        "sequana_quality_control": {"fn": "summary*.json"},
        "sequana_isoseq_qc": {"fn": "sequana_summary*.json"},
        "sequana_isoseq": {"fn": "sequana_summary*.json"},
        "sequana_kraken": {"fn": "sequana_kraken_summary.json"},
        "sequana_pacbio_amplicon": {"fn": "sequana_pacbio_amplicon_*.json"},
        "sequana_bamtools_stats": {"fn": "sequana_bamtools_stats_*.txt"},
    }
    config.update_dict(config.sp, my_search_patterns)



from multiqc.utils import config


# Add search patterns and config options for the things that are used in
# MultiQC_sequana
def multiqc_sequana_config():
    """Set up MultiQC config defaults for this package"""
    sequana_search_patterns = {
        "sequana/pacbio_qc": {"fn": "summary_*.json"},
        "sequana/quality_control": {"fn": "summary.json"},
        "sequana/coverage": {
            "fn": "sequana_summary_coverage.json",
        },
        "sequana/isoseq": {
            "fn": "sequana_summary_isoseq.json",
        },
        "sequana/isoseq_qc": {
            "fn": "sequana_summary_isoseq_qc.json",
        },
        "sequana/pacbio_amplicon": {
            "fn": "sequana_pacbio_amplicon*.json",
        },
    }
    config.update_dict(config.sp, sequana_search_patterns)
    # config.fn_clean_exts.append({'type': 'regex', 'pattern': 'summary_*.*'})
    config.update_dict(
        config.table_columns_visible,
        {
            "FastQC": {
                "percent_duplicates": False,
                "total_sequences": False,
            },
            "Samtools Stats": {
                "non-primary_alignments": False,
                "reads_mapped": False,
                "reads_mapped_percent": False,
                "raw_total_sequences": False,
            },
        },
    )

    # config.log_filesize_limit =  500000
    # in sequana_coverage, large html files are created.
    # they can be ignored
    config.fn_ignore_files = ["*html"]

    # config.update_dict(config.log_filesize_limit, 500000)



#
#  This file is part of Damona software
#
#  Copyright (c) 2020 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Tools to install images/binaries"""
import os
import pathlib
import shutil
import subprocess
import sys
import time

from easydev import md5
from spython.main import Client

from damona import Environ, Registry
from damona import version as damona_version
from damona.common import ImageReader, requires_singularity
from damona.registry import Software
from damona.utils import download_with_progress

DAMONA_PATH = os.environ["DAMONA_PATH"]


import colorlog

logger = colorlog.getLogger(__name__)


__all__ = ["LocalImageInstaller", "RemoteImageInstaller", "BinaryInstaller"]


class CMD:
    def __init__(self, cmd):
        self.cmd = cmd
        self.version = damona_version

    def __repr__(self):
        if self.cmd[0].endswith("damona"):
            cmd = "damona " + " ".join(self.cmd[1:])
            return cmd
        else:  # not a correct damona command (maybe from pytest)
            return "#test"


class ImageInstaller:
    @requires_singularity
    def _are_binaries_findable(self):
        # TODO add sanity check that stops the installation if a failure occurs
        for binary in self.binaries:
            # is it found in the image ? without -v may return error code
            # so we first test with --version option but some tools will use -v
            # so we test --version and -v cases. Finally, if there is no
            # version, let us just try without arguments. Most probably an error
            # code is returned. ut if not found at all, a 'executable file not
            # found' message should be returned by bash. Not ideal but will do
            # for now.
            cmd1 = f"singularity exec {self.input_image.filename} {binary}"
            cmd2 = f"singularity exec {self.input_image.filename} {binary} -v"
            cmd3 = f"singularity exec {self.input_image.filename} {binary} --version"
            # ideally we should use commad but not present on all systems....
            # cmd = f"singularity exec {self.input_image.filename} command -v {binary}"

            status = subprocess.Popen(cmd3, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            status.wait()
            if status.returncode == 0:
                logger.info(f"'{binary}' binary found in the container. Planned to be installed")
                continue

            status = subprocess.Popen(cmd2, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            status.wait()
            if status.returncode == 0:  # pragma: no cover
                logger.info(f"'{binary}' binary found in the container. Planned to be installed")
                continue

            status = subprocess.Popen(cmd1, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            status.wait()
            if status.returncode == 0:  # pragma: no cover
                logger.info(f"'{binary}' binary found in the container. Planned to be installed")
                continue
            else:
                if status.stderr is None:
                    logger.critical("Unknown error due to errocode != 0 but not error message reported")
                    # FIXME. not sure this ever happens
                    return False

                error = status.stderr.read()
                if "executable file not found" in error.decode():
                    logger.critical(f"{binary} executable not available in the image {self.input_image.filename}")
                    return False
                else:
                    # Other type of error could simply be help printed to stderr...
                    # let us assume this is the case but let us print a warning
                    # for future debugging
                    logger.warning(
                        f"Error while testing presence of {binary}. Most probably related to missing arguments. Here is the error code returned: \n\n{error}\n\n Proceed"
                    )
        return True

    def install_binaries(self, force=False):
        if self.image_installed:
            bininst = BinaryInstaller(self.binaries, self.input_image.filename)
            bininst.install_binaries(force=force)
        else:
            logger.critical("The container image has not been installed. So, binaries cannot be installed either")


class LocalImageInstaller(ImageInstaller):
    """Install a local singularity image.


    When using the standalone, you can install a local singularity even though it is not registered.
    This is not recommended but very useful for developers::

        damona install fastqc_0.11.9.img --binaries fastqc

    Since there is no registry, you can set the list of binaries.

    """

    def __init__(self, image_name, cmd=None, binaries=None):
        """.. rubric:: **Constructor**

        :param str image_name: The location of the singularity image to be installed.
        :param cmd: internal place holder to fill the history log with the calling command.
        :param list binaries: The list of binaries to be installed and expected to be
            found in the :attr:`image_name`

        """
        # must be a valid name with version, name, etc so that executable name
        # can be guessed
        super(LocalImageInstaller, self).__init__()

        #if os.path.exists(image_name) and pathlib.Path(image_name).is_dir():
        #    logger.error(f"image name must be a singulatity file, not a directory")
        #    sys.exit(1)

        self.input_image = ImageReader(image_name)
        self.target = pathlib.Path(DAMONA_PATH) / "images" / self.input_image.shortname
        self.images_directory = pathlib.Path(DAMONA_PATH) / "images"
        self.cmd = cmd

        # If no binaries are provided, we will guess from the name of the image
        # the unique possible binary to be found in the image.
        if binaries is None:
            self.binaries = [self.input_image.guessed_executable]
        else:
            self.binaries = binaries

        self.image_installed = False

    def is_valid(self):
        """Check whether binaries are findable in the image"""

        # let us check the  presence of the binaries in the image
        status = self._are_binaries_findable()
        return status

    def install_image(self, force=False):
        """Install the singularity image in the environment

        If an image exists with the same name, we check the md5sum of the destination and target.
        If they are identical, no need to overwrite the destination.
        """
        if (self.images_directory / self.input_image.shortname).exists():
            md5_target = md5(self.images_directory / self.input_image.shortname)
            if md5_target == self.input_image.md5:
                logger.info("Image with same md5 exists already. No need to copy")
                # image is not copied since it is already installed
                self.image_installed = True
            else:  # pragma: no cover
                if force:
                    self.copy()
                else:
                    logger.warning(
                        f"{self.input_image.filename} exists with different md5sum. Please use --force to overwrite"
                    )
        else:
            self.copy()

    def copy(self):  # pragma: no cover
        logger.info(f"Copying {self.input_image.filename} into {self.images_directory}")
        shutil.copy(self.input_image.filename, self.images_directory)
        with open(self.images_directory / "history.log", "a+") as fout:
            cmd = CMD(self.cmd).__repr__()
            fout.write(f"{time.asctime()}: {cmd}\n")
        # if the image has been copied properly, we set this flag to True so
        # that binaries can be installed
        self.image_installed = True


class RemoteImageInstaller(ImageInstaller):
    """Install a singulariy image and binaries from a damona registered image.

    The following command will download the fastqc singularity version 0.11.9
    from the damona website.::

        damona install fastqc:0.11.9

    It introspects all registry files to be found in damona package.

    However, if you have a damona registry on your website called registry.txt, you can download
    third-party images/software from any URL as follows::

        damona install fastqc:0.11.9 --url https://yourwebsite/.../registry.txt

    If you use this command often, you may want to add this line in your damona.cfg file. By
    default you should have this section in your config file::

        [urls]
        damona=https://biomics.pasteur.fr/salsa/damona/registry.txt

    so that you can simply type::

        damona install fastqc:0.11.9 --url damona

    """

    def __init__(self, image_name, binaries=None, from_url=None, cmd=None):
        """.. rubric:: **Constructor**

        :param str image_name: The location of the singularity image to be installed.
        :param list binaries: The list of binaries to be installed and expected to be
            found in the :attr:`image_name`
        :param cmd: internal place holder to fill the history log with the calling command.
        :param from_url: provide a URL if you want to use a third-party online registry

        """
        super(RemoteImageInstaller, self).__init__()

        self.image_name = image_name
        self.registry = Registry(from_url=from_url)
        self.images_directory = pathlib.Path(DAMONA_PATH) / "images"

        # if alias is given, this attribute contains the augmented (real) URL
        # but it contains registry.txt so we need to get rid of the registry filename
        try:
            self.from_url = self.registry.from_url.replace("registry.txt", "")
        except AttributeError:  # the url may not be set
            self.from_url = from_url
        self.cmd = cmd
        self.binaries = binaries
        self.image_installed = False

    def is_valid(self):
        return True

    @requires_singularity
    def pull_image(self, output_name=None, force=False):
        """Pull and Install a singularity image from the web."""

        self.image_installed = False

        # the name must be valid. we use _ to separate name and version for the
        # singularity filemane but the image have a tag (version) and name
        # separated by a :

        # Our registry uses underscore. The output filename should use _ as well
        # However, for users and image stored on sylabs or singularity hub, we
        # use the character :

        # let us first retrieve the image and save it in a temporary file.
        # we keep the original name by only replacing the : with underscore so
        # that we can use ImageReader class to help us later on.

        if ":" in self.image_name:
            logger.info(f"Looking for {self.image_name}...")
            # e.g. fastqc:0.11.9
            registry_name = self.image_name  # .replace(":", "_")

            if self.image_name not in self.registry.registry.keys():
                logger.critical("invalid image name provided: {}. type 'damona list'".format(self.image_name))
                guess = self.image_name.split(":")[0][0:4]
                # here we reverse x to start from the end and replace the last _
                # by :
                guesses = [
                    x[::-1].replace("_", ":", 1)[::-1] for x in self.registry.registry.keys() if x.startswith(guess)
                ]
                logger.critical("Maybe you meant one of: {}".format(guesses))
                sys.exit(1)
        else:
            soft = Software(self.image_name)

            try:
                r = soft.releases
            except AttributeError:
                logger.error(f"Could not find {self.image_name}. Use 'damona search' maybe")
                sys.exit(1)
            latest = r.last_release
            logger.info(
                f"No version found after {self.image_name} (e.g. fastqc:0.11.8)."
                + f" Installing latest version {latest}"
            )
            registry_name = self.image_name
            # we look only at the prefix name, not the tag. so we should get the
            # registry names from self.registry that have the prefix in common,
            # then extract the versions, and figure out the most recent.

        registry_name = self.registry.find_candidate(registry_name)
        if registry_name is None:
            logger.critical(
                f"No image found for {self.image_name}. Make sure it is correct using 'damona search' command"
            )
            sys.exit(1)

        download_name = self.registry.registry[registry_name].download

        # get metadata from the registry recipe
        info = self.registry.registry[registry_name]
        logger.info(f"{info}")

        if self.binaries:
            pass
        elif info.binaries:
            self.binaries = info.binaries
        else:
            logger.warning(f"No binaries field found in registry of {registry_name}")
            self.binaries = [self.input_image.guessed_executable]

        # target file is stored in output_name variable
        if output_name is None:
            # original code
            # output_name = registry_name.replace(":", "_") + ".img"
            # In order to allow a registry to refer to another existing container, 
            # we need to retrieve the name of the container itself rather than
            # building one from the registry name
            output_name = download_name.split("/")[-1]

        # here we check whether the image or binaries are already present.
        # if md5 is already provided, and image exists, nothing to copy
        if info.md5sum:
            if os.path.exists(self.images_directory / output_name):
                md5_target = md5(self.images_directory / output_name)

                if md5_target == info.md5sum:
                    logger.info("Remote image and local image are identical, no need to download/pull again")
                    self.input_image = ImageReader(self.images_directory / output_name)
                    self.image_installed = True
                    return
        else:
            logger.warning(
                f"md5 field not found or not filled in {registry_name}."
                f"To be fixed in https://github.com/damona/damona/recipes/{self.image_name}/registry.yaml "
            )

        # now that we have the registry name, we can download the image
        logger.info("Downloading {}".format(download_name))

        pull_folder = self.images_directory / "damona_buffer"

        if self.from_url:
            # The client does not support external https link other than
            # docker, library, shub.
            cmd = f"singularity pull --dir {pull_folder} "
            if force:
                cmd += " --force "
            cmd += f"{self.from_url}/{download_name}"
            print(cmd)
            subprocess.call(cmd.split())
        else:

            if download_name.startswith("https://"):
                logger.info(f"downloading into {pull_folder} {output_name}")

                download_with_progress(download_name, filename=str(pull_folder / output_name))

            elif download_name.startswith("docker://"):  # docker has no extension .img/.sig
                output_name += ".img"
                output_name = output_name.replace(":v", "_v")
                Client.pull(str(download_name), name=output_name, pull_folder=pull_folder, force=force)
            else:  # use singularity
                Client.pull(str(download_name), name=output_name, pull_folder=pull_folder, force=force)
        logger.info(f"Downloaded {output_name} into {pull_folder}")

        # Read the image, checking everything is correct

        self.input_image = ImageReader(pull_folder / output_name)
        shortname = self.input_image.shortname
        try:
            logger.info(f"Copying into damona image directory: {self.images_directory}")
            self.input_image.filename.rename(self.images_directory / shortname)
            self.input_image.filename = pathlib.Path(self.images_directory / shortname)
        except FileNotFoundError:
            logger.warning("File not installed properly. Stopping")
            self.image_installed = False

        # check the md5 validity
        if info.md5sum:
            if info.md5sum != self.input_image.md5:
                logger.warning(
                    "MD5 of downloaded image does not match the expected md5 found in the "
                    f"registry of {registry_name}. The latter may be incorrect in damona and needs "
                    f"to be updated in https://github.com/damona/damona/recipes/{self.image_name}/registry.yaml "
                    "or the donwload was interrupted"
                )

        self.image_installed = True


class BiocontainersInstaller(ImageInstaller):
    """

    Using bioservices.Biocontainers, we can retrieve all tools and versions.
    We could store the list once in a while in damona/biocontainers/registry.yaml

        from bioservices import Biocontainers
        b = Biocontainers()
        info = b.get_tools()
        tools = {}
        for name, versions in zip(info['name'], info['versions']):
            tools[name] = [x['meta_version'] for x in versions]

    To find valid hit on biocontainer, one can use::

        damona searh hisat2 --include-biocontainers

    This does not currently work because it searches for a docker that is not working.
    damona install hisat2

    """

    def __init__(self, image_name, binaries=None, cmd=None):
        super(BiocontainersInstaller, self).__init__()

        # get the biocontainers prefix and the name:version suffix
        prefix, suffix = image_name.split("/")

        if prefix != "biocontainers":
            logger.error(f"Biocontainers name must start with 'biocontainers/' ({prefix} provided")
            sys.exit(1)

        try:
            name, version = suffix.split(":")
            self.name = name
            self.version = version
        except ValueError as err:
            logger.error(f"Biocontainers name must be formatted as NAME:VERSION ({suffix} provided)")
            sys.exit(1)

        self.registry = Registry(biocontainers=True)

        if suffix not in self.registry.registry:
            logger.error(
                f"{name} not found in the biocontainers registry. Use damona search PATTERN --include-biocontainers"
            )
            sys.exit(1)

        self.image_name = image_name
        self.images_directory = pathlib.Path(DAMONA_PATH) / "images"

        self.cmd = cmd
        self.binaries = binaries
        self.image_installed = False

    def is_valid(self):
        return True

    @requires_singularity
    def pull_image(self, output_name=None, force=False):
        """Pull and Install a biocontainer image."""
        self.image_installed = False

        download_name = self.image_name

        # now that we have the registry name, we can download the image

        pull_folder = self.images_directory / "damona_buffer"
        if output_name is None:
            output_name = self.image_name.replace("biocontainers/", "")
            output_name = output_name.replace(":", "_") + ".img"

        logger.info(f"Downloading {self.image_name} into {output_name}")

        #
        Client.pull(f"docker://{self.image_name}", name=output_name, pull_folder=pull_folder, force=force)
        logger.info(f"File {self.image_name} uploaded to {pull_folder} as {output_name}")

        #
        self.input_image = ImageReader(pull_folder / output_name)

        shortname = self.input_image.shortname
        try:
            logger.info(f"Copying into damona image directory: {self.images_directory}")
            self.input_image.filename.rename(self.images_directory / shortname)
            self.input_image.filename = pathlib.Path(self.images_directory / shortname)
        except FileNotFoundError:
            logger.warning("File not installed properly. Stopping")
            self.image_installed = False

        # for the install_binaries to work, we need to set the binaries
        prefix, suffix = self.image_name.split("/")
        self.binaries = self.registry.registry[suffix].binaries

        self.image_installed = True


class BinaryInstaller:
    """Install a binary in the bin/ directory of the current environment given its image"""

    def __init__(self, binaries, parent_image_path, env_name=None):
        """.. rubric:: **Constructor**

        :param list binaries: list of binaries to install
        :param str parent_image_path: the location of the images where binaries are to be found

        """
        #: instance of :class:`damona.common.ImageReader`
        self.image = ImageReader(parent_image_path)
        self.binaries = binaries

        if env_name:
            from damona.common import get_damona_path

            self.env_name = get_damona_path() / "envs" / env_name
        else:
            env = Environ()
            self.env_name = env.get_current_env()

    @requires_singularity
    def install_binaries(self, force=False):
        """Install an image and its binaries

        Given the :attr:`~damona.install.BinaryInstaller.image`, we install a set of :attr:`~damona.install.BinaryInstaller.binaries` to be found in the
        image registry. If we install a binary again, no need to rewrite the command. For example,
        imagine that you installed fastqc.0.11.9 for the first time, then in the commands
        here below::

            damona install fastqc:0.11.9
            damona install fastqc:0.11.9 --force

        the second command has no effect. If we now install a new version::

            damona install fastqc:0.11.8

        The previous binary (v0.11.9) will be replaced by the new one (v0.11.8).
        The old image is kept in the images directory (it may be used in another
        environment indeed).
        """

        bin_directory = self.env_name / "bin"
        logger.info(bin_directory)

        for binary in sorted(self.binaries):
            bin_path = pathlib.Path(bin_directory) / binary

            CMD = """singularity -s exec ${{DAMONA_SINGULARITY_OPTIONS}} {} {} ${{1+"$@"}}"""
            CMD = CMD.format(f"${{DAMONA_PATH}}/images/{self.image.shortname}", binary)

            if bin_path.exists() and force is False:
                name = pathlib.Path(bin_path).name
                path = str(bin_path).rstrip(bin_path.name)
                logger.warning(
                    f"Binary {binary} exists already in {path} and was not changed. Use --force to overwrite"
                )
            else:
                # create a new binary content and save into an executable file
                with open(bin_path, "w") as fout:
                    fout.write(f"#!/bin/sh\n{CMD}")

                # update permission so that it is executable
                os.chmod(bin_path, 0o755)

                # print message
                name = pathlib.Path(bin_path).name
                path = str(bin_path).rstrip(bin_path.name)
                logger.info(f"Created binary {name} in {path}")



#
#  This file is part of Damona software
#
#  Copyright (c) 2016 - Damona Development Team
#
#  File author(s):
#      Thomas Cokelaer <thomas.cokelaer@pasteur.fr>
#
#  Distributed under the terms of the 3-clause BSD license.
#  The full license is in the LICENSE file, distributed with this software.
#
#  website: https://github.com/cokelaer/damona
#  documentation: http://damona.readthedocs.io
#
##############################################################################
"""Builder for containers from docker or singularity images"""
import getpass
import os
import subprocess
import sys
import tempfile

from damona.common import Damona, requires_singularity

manager = Damona()


import colorlog

logger = colorlog.getLogger(__name__)

__all__ = ["Builder", "BuilderFromSingularityRecipe", "BuilderFromDocker"]


class Builder:
    """Build a container using different framework

    Input can be a singularity or docker container/recipes.

    """

    def __init__(self):
        # just to check that a username is defined before building the image
        self.username

    def get_temp_file(self):
        filename = tempfile.NamedTemporaryFile(dir=manager.config_path, suffix=".img")
        return filename

    def _get_username(self):
        return getpass.getuser()

    username = property(_get_username, doc="return the username (unix)")

    def teardown(self, dest):
        """finalise the build

        Currently, just print information
        """
        # once built and install, we can stop and save information in
        #  the history if it was susccessul
        logger.info(f"Image built in {dest}")


class BuilderFromDocker(Builder):
    """Install a singularity container image and executable from a docker hub container.

    This command creates automatically the destination bowtie2_2.4.1.img because the version
    follows Damona convention::

        damona build docker://biocontainers/bowtie2:v2.4.1_cv1

    This command will fail::

        damona build docker://alpine

    but this one will tell the version::

        damona build docker://alpine --destination alpine_1.0.0.img

    We assume that the name of the executable is the name of the container.
    User may provide a binary or list of binaries to install from the container if it is known.
    """

    def __init__(self):
        super(BuilderFromDocker, self).__init__()
        logger.info("Building a singularity image from docker")

    @requires_singularity
    def build(self, dockerhub_name, destination=None, force=False):
        """Build the singularity image from docker image"""

        # if the build is successful, we will copy the image
        # into the current environment.
        # Let us check now the present of the image and its executable
        if destination is None:  # pragma: no cover
            if ":v" in dockerhub_name:
                name, version = dockerhub_name.split(":v")
                name = name.split("/")[-1]
                destination = name + "_" + version + ".img"
            else:
                logger.error(
                    "The container has no version using the v: separator (e.g., bowtie2:v1.0.0). No destination was provided either; please use e.g. --destination NAME_X.Y.Z.img where X.Y.Z is te version"
                )
                sys.exit(1)
        else:
            if destination.endswith(".sif") is False and destination.endswith(".img") is False:
                logger.error("destination name must end in .sif or .img")
                sys.exit(1)

        if os.path.exists(destination):
            answer = None
            if force is True:
                answer = "yes"
            while answer not in ["yes", "no"]:
                answer = input(f"Image ({destination}) exists already in Damona, do you want to overwrite it ?")
                if answer == "no":
                    return
                elif answer == "yes":
                    pass
                else:
                    logger.error("please answer yes or no")
                    sys.exit(1)

        # build the image
        cmd = f"sudo singularity pull --force {destination} docker://{dockerhub_name} "
        logger.info(f"Running : {cmd}")
        subprocess.call(cmd.split())

        cmd = f"sudo chown {self.username}:{self.username} {destination}"
        subprocess.call(cmd.split())

        self.teardown(destination)


class BuilderFromSingularityRecipe(Builder):
    """Build a container from its singularity recipe

    This command creates the destination file bowtie2_2.4.1.img::

        damona build Singularity.bowtie2_2.4.1

    """

    def __init__(self):
        super(BuilderFromSingularityRecipe, self).__init__()
        logger.info("Building a Singularity image from a Singularity recipe")

    @requires_singularity
    def build(self, recipe, destination=None, force=False):
        if os.path.basename(recipe).startswith("Singularity.") is False:
            logger.error("Recipe must start with Singularity.")
            sys.exit(1)

        if destination is None:  # FIXME: do the same as for docker files ?
            destination = os.path.basename(recipe).replace("Singularity.", "") + ".img"

        if os.path.exists(destination):
            answer = None
            if force is True:
                answer = "yes"
            while answer not in ["yes", "no"]:
                answer = input(f"Image ({destination}) exists already, do you want to overwrite it ?")
                if answer == "no":
                    return
                elif answer == "yes":
                    pass
                else:
                    logger.error("please answer yes or no")
                    sys.exit(1)

        # build the image
        cmd = f"sudo singularity build --force {destination} {recipe} "
        logger.info(f"Running : {cmd}")
        status = subprocess.call(cmd.split())
        if status != 0:  # pragma: no cover
            logger.error("An error occured")
            sys.exit(1)

        cmd = f"sudo chown {self.username}:{self.username} {destination}"
        subprocess.call(cmd.split())

        self.teardown(destination)



#!/usr/bin/env python

""" MultiQC module to parse output from sequana"""
import os
import re

# prevent boring warning (version 1.0)
import logging

logging.captureWarnings(True)
from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph, table, heatmap, bargraph

logging.captureWarnings(False)

# Initialise the logger
import colorlog

logger = colorlog.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Sequana/pacbio",  # name that appears at the top
            anchor="sequana",  # ??
            target="sequana",  # Name show that link to the following href
            href="http://github.com/sequana/sequana/",
            info="pipelines multi Summary",
        )

        self.sequana_data = {}
        for myfile in self.find_log_files("sequana_pacbio_qc"):
            logging.info("Parsing {}".format(myfile))
            # print( myfile['f'] )       # File contents
            # print( myfile['s_name'] )  # Sample name (from cleaned filename)
            # print( myfile['fn'] )      # Filename
            # print( myfile['root'] )    # Directory file was in
            name = myfile["s_name"]
            if name.startswith("sequana_summary_pacbio_qc_"):
                name = name.replace("sequana_summary_pacbio_qc_", "")
            self.sequana_data[name] = self.parse_logs(myfile["f"])

        if len(self.sequana_data) == 0:
            logger.debug("No samples found: sequana_pacbio_qc")
            raise UserWarning

        info = "<ul>"
        for this in sorted(self.sequana_data.keys()):
            info += '<li><a href="../{}/summary.html">{}</a></li>'.format(this, this)
        info += "</ul>"
        href = "http://sequana.readthedocs.io/en/main/"
        target = "Sequana"
        mname = '<a href="{}" target="_blank">{}</a> individual report pages:'.format(href, target)
        self.intro = "<p>{} {}</p>".format(mname, info)

        logger.info("Found {} reports".format(len(self.sequana_data)))

        self.populate_columns()
        self.add_count_section()
        self.add_read_length_section()
        self.add_hist_GC()
        self.add_hist_length()
        # Add histogram GC, SNR, read length

    def add_read_length_section(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"mean": self.sequana_data[name]["mean_length"]}

        pconfig = {
            "title": "Mean read length",
            "percentages": False,
            "min": 100,
            "logswitch": True,
        }

        self.add_section(
            name="Mean read length",
            anchor="mean_read_length",
            description="Mean length of the reads per sample",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_count_section(self):
        data = {}
        for name in self.sequana_data.keys():
            data[name] = {"count": self.sequana_data[name]["count"]}

        pconfig = {
            "title": "Number of reads per sample",
            "plot_tt_percentages": False,
            "min": 100,
            "logswitch": True,
        }

        self.add_section(
            name="Number of reads",
            anchor="number_of_reads",
            description="Number of reads per sample.",
            helptext="",
            plot=bargraph.plot(data, None, pconfig),
        )

    def add_hist_GC(self):
        """Create the HTML for the FastQC GC content plot"""
        data = dict()
        data_norm = dict()
        for s_name in self.sequana_data:
            try:
                X = self.sequana_data[s_name]["hist_gc"]["X"]
                Y = self.sequana_data[s_name]["hist_gc"]["Y"]
                Y = [y / float(sum(Y)) for y in Y]
                data[s_name] = {x: 10 * y for x, y in zip(X[1:], Y)}
            except KeyError:
                pass
            # else:
            #    data_norm[s_name] = dict()
            #    total = sum( [ c for c in data[s_name].values() ] )
            #    for gc, count in data[s_name].items():
            #        data_norm[s_name][gc] = (count / total) * 100

        if len(data) == 0:
            logger.debug("no data for the GC content plots")
            return None

        pconfig = {
            "id": "sequana_pacbio_per_sequence_gc_content_plot",
            "title": "Per Sequence GC Content",
            "ylab": "Count",
            "xlab": "% GC",
            "ymin": 0,
            #'ymax': 0.2,
            "xmax": 100,
            "xmin": 0,
            "yDecimals": False,
            "tt_label": "<b>{point.x}% GC</b>: {point.y}",
            #'colors': self.get_status_cols('per_sequence_gc_content'),
            "data_labels": [{"name": "Percentages", "ylab": "Percentage"}, {"name": "Counts", "ylab": "PDF"}],
        }

        self.add_section(
            name="Per Sequence GC Content",
            anchor="fastqc_per_sequence_gc_content",
            description="GC content (normalised)",
            # plot = linegraph.plot([data_norm, data], pconfig))
            plot=linegraph.plot(data, pconfig),
        )

    def add_hist_length(self):
        """Create the HTML for the FastQC GC content plot"""
        data = dict()
        data_norm = dict()
        for s_name in self.sequana_data:
            X = self.sequana_data[s_name]["hist_read_length"]["X"]
            Y = self.sequana_data[s_name]["hist_read_length"]["Y"]
            # Y = [y / sum(Y) for y in Y]
            data[s_name] = {x: y for x, y in zip(X[1:], Y)}
            try:
                X = self.sequana_data[s_name]["hist_read_length"]["X"]
                Y = self.sequana_data[s_name]["hist_read_length"]["Y"]
                # Y = [y / sum(Y) for y in Y]
                data[s_name] = {x: y for x, y in zip(X[1:], Y)}
            except KeyError:
                pass

        if len(data) == 0:
            logger.debug("no data for the read length plots")
            return None

        pconfig = {
            "id": "sequana_pacbio_hist_length",
            "title": "Read length",
            "ylab": "#",
            "xlab": "Length",
            "ymin": 0,
            "xmax": 50000,
            "xmin": 0,
            "yDecimals": False,
            "tt_label": "<b>{point.x}Length</b>: {point.y}",
            #'colors': self.get_status_cols('per_sequence_gc_content'),
            #'data_labels': [
            #    {'name': 'Percentages', 'ylab': 'Percentage'},
            #    {'name': 'length', 'ylab': '#'}
            # ]
        }

        self.add_section(
            name="Read length histograms",
            anchor="fastqc_per_sequence_gc_content",
            description="Read length histogram",
            plot=linegraph.plot(data, pconfig),
        )

    def parse_logs(self, log_dict):
        import json

        log_dict = json.loads(log_dict)
        data = {}
        data["count"] = log_dict["read_stats"]["count"]
        data["mean_length"] = log_dict["read_stats"]["mean"]
        data["mean_gc"] = log_dict["mean_gc"]
        data["hist_gc"] = log_dict["hist_gc"]
        data["hist_read_length"] = log_dict["hist_read_length"]
        return data

    def populate_columns(self):
        headers = {}
        if any(["count" in self.sequana_data[s] for s in self.sequana_data]):
            headers["count"] = {
                "title": "Read count",
                "description": "read count",
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,0d}",
                "shared_key": "count",
            }

        if any(["mean_gc" in self.sequana_data[s] for s in self.sequana_data]):
            headers["mean_gc"] = {
                "title": "Mean GC (%)",
                "description": "mean GC content in %",
                "max": 100,
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.2f}",
            }

        if any(["mean_length" in self.sequana_data[s] for s in self.sequana_data]):
            headers["mean_length"] = {
                "title": "Mean length",
                "description": "mean length",
                "max": 15000,
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.2f}",
            }
        if len(headers.keys()):
            self.general_stats_addcols(self.sequana_data, headers)
